// implements the newcamd protocol (versions 5.20 & 5.25)

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/poll.h>

#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "newcamd.h"
#include "des.h"
#include "newcamd-client.h"


// ------------------------------- newcamd client class member implementation ------------------------------

newcamd_client::newcamd_client(char *server_url, unsigned int timeout, protocol_version_t version)
{
	set_protocol_version( version );
	
	status = 0;
	status_description[0] = 0;
	ready = false;
	network_timeout = timeout;
	
	sent_sequence_id = 0;
	received_sequence_id = 0;

	memset(cw_0, 0, sizeof(cw_0));
	memset(cw_1, 0, sizeof(cw_1));
	
	strcpy(url, server_url);
	if (strncmp(url,"N: ",3) != 0) return;
	memset(username, 0, sizeof(username));
	memset(password, 0, sizeof(password));
	memset(hostname, 0, sizeof(hostname));
	port = 0;
	fd=-1;	
	user_id = -1;
	card_caid = 0;
	card_provider_id_count = 0;
	memset(cam_id, 0, sizeof(cam_id));
	
	memset(random_key, 0, sizeof(random_key));
	memset(des_key, 0, sizeof(des_key));
	memset(login_key, 0, sizeof(login_key));
	memset(session_key, 0, sizeof(session_key));
	
	memset(expanded_login_key, 0, sizeof(expanded_login_key));
	memset(expanded_session_key, 0, sizeof(expanded_session_key));
	
	parse_url();
	server_connect();
	if( status ) return;
	server_login();
	if( status ) return;
	ready = true;
	dbg = 0;
}


bool newcamd_client::is_ready( void )
{
	return ready;
}

void newcamd_client::set_protocol_version( protocol_version_t version )
{
	protocol_version = version;
	switch( version )
	{
		case VERSION_520:
			custom_data_length = 4;
			break;
			
		case VERSION_525:
			custom_data_length = 8;
			break;
		
		default:
			custom_data_length = 8;
			break;
	}
	
	set_sid( 0x6d67 );    // x0666   //0x8888 Oscam
}

void newcamd_client::set_sid( unsigned short int sid )
{
	memset(custom_data, 0, sizeof(custom_data));
	switch( protocol_version )
	{
		case VERSION_520:
			custom_data[2] = sid >> 8;
			custom_data[3] = sid & 0xFF;
			break;
			
		case VERSION_525:
			custom_data[0] = sid >> 8;
			custom_data[1] = sid & 0xFF;
			break;
		
		default:
			custom_data[0] = sid >> 8;
			custom_data[1] = sid & 0xFF;
			break;
	}
}

void newcamd_client::parse_url( void )
{
	char config_key[29];
	char emm_config[4];
	unsigned char config_key_byte_count = sizeof(des_key);
	int des_key_byte;
	
	memset(config_key, 0, sizeof(config_key));
	sscanf(url, "N: %s %d %s %s %s %s", hostname, &port, username, password, config_key, emm_config);

		printf("[Newcamd] User : %s : %s\n", username, password);
		//printf("[Newcamd] Conf : %s : %s\n", config_key, emm_config);
		printf("[Newcamd] Host : %s : %d\n", hostname, port);
		
	while( config_key_byte_count-- )
	{
		sscanf(&config_key[config_key_byte_count * 2], "%02X", &des_key_byte);
		des_key[config_key_byte_count] = (unsigned char)des_key_byte;
		//printf("key %d : %d\n",config_key_byte_count,des_key_byte);
	}
}
	
newcamd_client::~newcamd_client( void )
{
	server_disconnect();
}

void newcamd_client::server_connect( void )
{
	struct hostent *host_address;
	struct sockaddr_in socket_address;
	
	host_address = gethostbyname(hostname);
	if(!host_address)
	{
		status = 1;
		printf("[Newcamd] Could not resolve hostname.\n");
		return;
	}
	
	if((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		status = 1;
		printf("[Newcamd] Could not get socket descriptor.\n");
		return;
	}
	
    socket_address.sin_family = AF_INET;
    socket_address.sin_port = htons(port);
    socket_address.sin_addr.s_addr = ((struct in_addr *) host_address->h_addr)->s_addr;
	
	if( connect(fd, (struct sockaddr *)(&socket_address), sizeof(struct sockaddr)) < 0 )
	{
		status = 1;
		printf("[Newcamd] Failed to connect to remote server.\n");
		return;
	}
	
	status = 0;
	printf("[Newcamd] Successfully connected to server.\n");
}

void newcamd_client::server_disconnect( void )
{       if (fd>=0)
		{
		if (shutdown(fd, 2)<0) perror("socket close");
		if (close(fd)<0) perror("socket close");
		}
}

void newcamd_client::server_login( void )
{
	unsigned char buffer[CWS_NETMSGSIZE];
	memset(buffer, 0, CWS_NETMSGSIZE);
	
	// ----------------------------------------- LOGIN --------------------------------------------------
	//server should initiate sequence with 14-byte random key. get it and create the login key
	if( network_read(random_key, sizeof(random_key)) != sizeof(random_key) )
	{
		status = 1;
		printf( "[Newcamd] Failed to receive random server key.");
		return;
	}
	create_login_key(); //xor random key with common des key
	
	//build login command data
	char *encrypted_password = crypt(password,"$1$abcdefgh$");
	unsigned short int username_length = strlen(username);
	unsigned short int password_length = strlen(encrypted_password);
  	memcpy(buffer, username, username_length + 1);
  	memcpy(buffer + username_length + 1, encrypted_password, password_length + 1);
  	
  	//send login command, if error, exit.
  	send_command(MSG_CLIENT_2_SERVER_LOGIN, buffer, username_length + 1 + password_length + 1, custom_data, LOGIN_KEY);
  	if( status ) return;
  	
  	//command sent successfully, get server response, if error, exit.	
  	int command = receive_command( NULL, LOGIN_KEY );
  	if( status ) return;
  	
  	//server response received succesfully, check if login is acknowledged, if not, exit.
  	if( command != MSG_CLIENT_2_SERVER_LOGIN_ACK )
  	{
  		status = 1;
  		printf("[Newcamd] Login request was rejected.\n");
  		return;
  	}
  	
  	// --------------------------------------- CARD DATA REQUEST -------------------------------------
  	//login was successful, now create session key
  	create_session_key( encrypted_password ); //xor password with common des key
  	
  	//from now on, all messages are encrypted with the session key
  	//request card data
  	send_command(MSG_CARD_DATA_REQ, 0, 0, custom_data, SESSION_KEY);
  	if( status ) return;
  	
  	//command sent successfully, get server response, if error, exit.	
  	cardatalen=receive_message(buffer, NULL, SESSION_KEY );
  	if( status ) return;
  	memcpy(cardatareq,buffer+3,cardatalen-3);		//cardatalen and cdr is stroed to be used if needed. NCD to  N forwarding
//	debug_output(cardatareq,cardatalen-3,"CDR : ");
  	//server response received succesfully, check if login is acknowledged, if not, exit.
  	if( buffer[0] != MSG_CARD_DATA )
  	{
  		status = 1;
  		printf( "[Newcamd] Card data request failed.\n");
  		return;
  	}
  	

  	//parse card data
  	card_caid = buffer[4] << 8 | buffer[5]; //get caid. it is stored at bytes 4 & 5 (big-endian)
  	user_id = buffer[3]; //get user id assigned by server.
	printf("[Newcamd] Card ID : %04X,  Server User : %d\n",card_caid,user_id);
  	
  	//check if cam id was provided.
  	if ( memcmp(buffer + 6, "\0\0\0\0\0\0\0\0", 8) )
  	{
  		//cam id info is present, get the data according to caid
  		switch( card_caid >> 8 )
  		{
  			case 0x18: //nagra
  				{
				memcpy(cam_id, buffer + 6 + 4, 4);
				printf("[Newcamd] CAM ID Nagra\n");
  				break;
				}
			case 0x0d: //cw
				{  				
				memcpy(cam_id, buffer + 6 + 3, 4);
				printf("[Newcamd] CAM ID CW\n");
  				break;
				}
			case 0x06: //irdeto
				{  				
				memcpy(cam_id, buffer + 6 + 5, 3);
				printf("[Newcamd] CAM ID Irdeto\n");
  				break;
				}
  		}
  	}
  	
  	card_provider_id_count = buffer[14] > MAX_CARD_PROVIDER_COUNT ? MAX_CARD_PROVIDER_COUNT : buffer[14];
	//fill provider id's collection
	printf ("[Newcamd] Idents  : ");
  	for(int i=0;i<card_provider_id_count;i++)
  		{
		card_provider_id[i] =  (buffer[15 + (i * 11)] << 16) | (buffer[16 + (i * 11)] << 8) | buffer[17 + (i * 11)];
		printf ("%06X ",card_provider_id[i]);
		}
	printf ("\n");

}

void newcamd_client::send_ecm(unsigned char *ecm_data, unsigned short int ecm_data_length, unsigned short int sid)
{
	unsigned char buffer[CWS_NETMSGSIZE];
	int bytes_received, i;
	
	//set the sid in custom data field
	set_sid( sid );
	
	//send the ecm, exit if error status
	send_message(ecm_data, ecm_data_length, custom_data, SESSION_KEY);
	if( status ) return;
	
	//this receive loop is to discard any old messages left in the socket read buffer due
	//to fast service switching (without this, fast service switching does not work properly).
	while( 1 )
	{
		bytes_received = receive_message(buffer, 0, SESSION_KEY);
		if( status ) return;

		if( received_sequence_id == sent_sequence_id ) break;

		if( received_sequence_id > sent_sequence_id )
		{
			status = 1;
			printf("[Newcamd] Future sequence id received. ECM request failed.\n");
			return;
		}

		/*
		if( received_sequence_id < sent_sequence_id )
			printf("seq id mismatch! sent: %d, recv: %d", sent_sequence_id, received_sequence_id);		
		*/
	}
	memcpy(echobuf,buffer,bytes_received);
	echobufsize=bytes_received;
	switch( bytes_received )
	{
		case 3 + 16: //19 bytes where received (3-byte header + 16-byte control words)
			//copy control words
			if( memcmp(buffer + 3 + 0, "\0\0\0\0", 4) ) memcpy(cw_0, buffer + 3 + 0, sizeof(cw_0));
			if( memcmp(buffer + 3 + 8, "\0\0\0\0", 4) ) memcpy(cw_1, buffer + 3 + 8, sizeof(cw_1));
			status = 0;
			if (dbg) 
			{
				printf( "[Newcamd] Control words received successfully.\n");
				for (i=0;i<19;i++) printf("%02x ",buffer[i]); printf("\n");
			}
			break;
			
		case 3: //only header received, no control words
			status = 1;
			if (dbg) printf( "[Newcamd] Control words not received.\n");
			break;
			
		default:
			status = 1;
			if (dbg) printf( "[Newcamd] Invalid server response (%d bytes received).\n", bytes_received);
			break;
			break;
	}
}

void newcamd_client::send_emm(unsigned char *emm_data, unsigned short int emm_data_length)
{
	unsigned char buffer[CWS_NETMSGSIZE];
	unsigned int bytes_received;
	

		send_message(emm_data, emm_data_length, 0, SESSION_KEY);
		if( status ) return;
			
		//this receive loop is to discard any old messages left in the socket read buffer due
		//to fast service switching (without this, fast service switching does not work properly).
		while( 1 )
		{
			bytes_received = receive_message(buffer, 0, SESSION_KEY);
			if( status ) return;
			
			if( received_sequence_id > sent_sequence_id )
			{
				status = 1;
				printf( "[Newcamd] Future sequence id received. EMM request failed.\n");
				return;
			}
			if( received_sequence_id == sent_sequence_id ) break;
		}
		
		if( bytes_received < 3 )
		{
			status = 1;
			if (dbg) printf( "[Newcamd] Invalid server response while processing EMM.\n");
		}
		else
		{
			if( buffer[1] & 0x10 )
			{
				status = 0;
				if (dbg) printf( "[Newcamd] EMM processed correctly by card.\n");
			}
			else
			{
				status = 0;
				if (dbg) printf( "[Newcamd] EMM rejected by card.\n");
			}
		}
	memcpy(echobuf,buffer,bytes_received);
	echobufsize=bytes_received;
}


void newcamd_client::sendack()
{
send_command(MSG_KEEPALIVE, 0, 0, NULL, SESSION_KEY);
}

void newcamd_client::send_command(net_msg_type_t command, unsigned char *data, unsigned short int data_length, unsigned char *custom_data, key_encryption_t key_type)
{
	unsigned char command_buffer[3 + data_length];
	command_buffer[0] = command;
	command_buffer[1] = (data_length >> 8) & 0x0F;
	command_buffer[2] = data_length & 0xFF;
	if( data ) memcpy(command_buffer + 3, data, data_length);
	
	return send_message(command_buffer, sizeof(command_buffer), custom_data, key_type);
}

net_msg_type_t newcamd_client::receive_command( unsigned char *custom_data, key_encryption_t key_type )
{
	unsigned char command_buffer[CWS_NETMSGSIZE];
	
	if( receive_message(command_buffer, custom_data, key_type) == 3 )
		return (net_msg_type_t) command_buffer[0];
	else
		return (net_msg_type_t)(-1);
}

int newcamd_client::receive_message(unsigned char *message, unsigned char *custom_data, key_encryption_t key_type)
{
	unsigned char message_buffer[CWS_NETMSGSIZE];
	unsigned short int message_length;
	int bytes_received;
	
	//message header length is 2-byte size header (unencrypted) + 2-byte sequence id (encrypted) + 
	//8 or 4-byte custom data field (encrypted).
	unsigned short int message_header_length = 2 + 2 + custom_data_length;
	
	//read first 2 bytes, this is the message length
	if( (bytes_received = network_read(message_buffer, 2)) != 2 )
	{
		switch( bytes_received )
		{
			case -1:
				status = 3;
				ready=false;
				printf( "[Newcamd] Failed to read network stream. Server disconnected.\n");
				return 0;
			
			case -2:
				status = 4;		//timeout 1 olabilir.
				ready=false;
				printf( "[Newcamd] Failed to read network stream. Time-out.\n");
				return 0;
				
			default:
				status = 5;
				ready=false;
				printf( "[Newcamd] Failed to read network stream.\n");
				return 0;
		}
	}
	
	//message length is big-endian
	message_length = message_buffer[0] << 8 | message_buffer[1];
	
	//check if message length is valid
	if( message_length > CWS_NETMSGSIZE - 2 )
	{
		status = 1;
		printf( "[Newcamd] Received message too large (%d bytes).\n", message_length + 2);
		return 0;
	}
	
	//read the rest of the message
	if( (bytes_received = network_read(message_buffer + 2, message_length)) != message_length )
	{
		switch( bytes_received )
		{
			case -1:
				status = 3;
				ready=false;
				printf( "[Newcamd] Failed to read network stream. Server disconnected.\n");
				return 0;
			
			case -2:
				status = 4;   //timeout 1 olabilir.
				ready=false;
				printf( "[Newcamd] Failed to read network stream. Time-out.\n");
				return 0;
				
			default:
				status = 5;
				ready=false;
				printf( "[Newcamd] Failed to read network stream.\n");
				return 0;
		}
	}
	message_length += 2;
	
	//decrypt the whole message, includes the 2-byte size header, using apropriate key_type	switch( key_type )
	switch( key_type )
	{
		case LOGIN_KEY:
			memcpy(des.des_key, expanded_login_key, sizeof(des.des_key));
			break;
			
		case SESSION_KEY:
			memcpy(des.des_key, expanded_session_key, sizeof(des.des_key));
			break;
	}
	des.decrypt(message_buffer, message_length);
	message_length -= sizeof(i_DES_cblock);
	
	//check sum the result
	if(des.xor_checksum(message_buffer + 2, message_length - 2))
	{
		status = 1;
		printf( "[Newcamd] Checksum error on decrypted packet.\n");
		return 0;
	}
	
	//get size from 12-bit command data length of message, add 3 bytes (from command header) to get 
	//total size of data to be returned
	unsigned short command_length = 
		((message_buffer[message_header_length + 1] & 0x0F) | message_buffer[message_header_length + 2]) + 3;
	
	//get the received sequence id (used for message synchronization)
	received_sequence_id = (message_buffer[2] << 8) | message_buffer[3];
	
	//if custom_data pointer is present, copy custom data from mesage to it
	if( custom_data ) memcpy(custom_data, message_buffer + 4, custom_data_length);
	
	memcpy(message, message_buffer + message_header_length, command_length);
	//debug_output(message_buffer, message_length, "< ");
	return command_length;
}

void newcamd_client::send_message(unsigned char *message, unsigned short int length, unsigned char *custom_data, key_encryption_t key_type)
{
	unsigned char message_buffer[CWS_NETMSGSIZE];
	memset(message_buffer, 0, CWS_NETMSGSIZE);
	
	//message header length is 2-byte size header (unencrypted) + 2-byte sequence id (encrypted) + 
	//8 or 4-byte custom data field (encrypted).
	unsigned short int message_header_length = 2 + 2 + custom_data_length;
	
	//total message length is the message header length + the length of the command data received.
	unsigned short int message_length = message_header_length + length;
	
	if( length < 3 || (message_header_length + length ) > CWS_NETMSGSIZE )
	{
		status = 1;
		printf( "[Newcamd] Invalid message size.\n");
		return;
	}
	
	//increment and big-endian set new session id to bytes 2 & 3 of message buffer
	message_buffer[2] = ((sent_sequence_id + 1) >> 8);
	message_buffer[3] = (sent_sequence_id + 1) & 0xFF;
	memcpy(message_buffer + message_header_length, message, length);  ///my fix
	
	//if custom data is received, embed it in the custom data section of the message
	//depending on protocol version, custom data length can be either 4 bytes or 8 bytes.
	if( custom_data ) memcpy(message_buffer + 4, custom_data, custom_data_length);
	
	//debug_output(message_buffer, message_length, "> ");
	
	//pad message for encrypting
	message_length = des.pad_message(message_buffer, message_length);
	
	//3des encrypt full message with appropriate key
	switch( key_type )
	{
		case LOGIN_KEY:
			memcpy(des.des_key, expanded_login_key, sizeof(des.des_key));
			break;
			
		case SESSION_KEY:
			memcpy(des.des_key, expanded_session_key, sizeof(des.des_key));
			break;
	}
	des.schedule_key();
	des.encrypt(message_buffer, message_length, message_buffer);
	
	//compensate for message padding (add 1 DES block to message length)
	message_length += sizeof(i_DES_cblock);
	
  	//set total message length UNENCRYPTED in bytes 0 & 1 of message
  	message_buffer[0] = (message_length - 2) >> 8;
  	message_buffer[1] = (message_length - 2) & 0xFF;
  	
  	//write message to network socket
  	if( network_write(message_buffer, message_length) != message_length )
  	{
  		status = 2;
		ready=false;
  		printf( "[Newcamd] Error writing network stream.\n");
  		return;
  	}
	sent_sequence_id++;
	
  	status = 0;
  	if (dbg) printf( "[Newcamd] Message sent successfully.\n");
}

void newcamd_client::create_login_key()
{
	//xors the received 14-byte random key with the 14-byte common key.
	//expands the resulting 14-byte login_key to a 16-byte expanded_login_key
	des.xor_buffers(login_key, random_key, des_key, sizeof(login_key));
	des.expand(login_key, expanded_login_key);
}

void newcamd_client::create_session_key( char *encrypted_password )
{
	//xors the common des key with the authenticaded user password to form the 14-byte session key
	//expands 14-byte sesion_key to a 16-byte expanded_session_key
	memcpy(session_key, des_key, sizeof(session_key));
	for(unsigned int i=0;i<strlen(encrypted_password);i++)
		session_key[i % sizeof(session_key)] ^= encrypted_password[i];
	des.expand(session_key, expanded_session_key);
}


int newcamd_client::network_write(unsigned char *buffer, unsigned int length)
{
	//write with time-out
	struct pollfd pfd;
	int write_result;
	
	pfd.fd = fd;
	pfd.events = POLLOUT | POLLERR;
	pfd.revents = 0;
	
	if(poll(&pfd, 1, network_timeout) != 1) return -2;
	
	if(pfd.revents & POLLOUT)
	{
		if( (write_result = write(fd, buffer, length)) <= 0 ) ready = false;
		//debug_output(buffer, write_result, "e> ");
		return write_result;
	}
	else
		return -1;
}

int newcamd_client::network_read(unsigned char *buffer, unsigned int length, int override_timeout)
{
	//read with time-out
	struct pollfd pfd;
	int read_result;
	
	pfd.fd = fd;
	pfd.events = POLLIN | POLLERR;
	pfd.revents = 0;
	
	if( override_timeout < 0 ) override_timeout = network_timeout;
	
	if(poll(&pfd, 1, override_timeout) != 1) return -2;
	
	if(pfd.revents & POLLIN)
	{
		if( (read_result = read(fd, buffer, length)) <= 0 ) ready = false;
		//debug_output(buffer, read_result, "e< ");
		return read_result;
	}
	else
		return -1;
}



void newcamd_client::debug_output(unsigned char *buffer, unsigned int length, char *header)
{
	printf("%s ", header);
	for(unsigned int i=0;i<length;i++) printf("%02X ", buffer[i]);
	printf("\n");
}

